/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _XOPEN_SOURCE 500
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>

int size_in_blocks(int fd)
{
  struct stat stbuf;
  fstat(fd,&stbuf);
  if (stbuf.st_size % 512) return 0;
  return stbuf.st_size/512;
}

int store_fd;

void get_block(uint32_t n, uint8_t *buf)
{
  if (n)
    assert(pread(store_fd,buf,512,512*n)==512);
  else
    memset(buf,0,512);
}

uint32_t pointers[4][128];

int main(int argc, char ** argv)
{
  uint32_t store_size;
  uint8_t buf[512];
  uint32_t idx,nidx;
  uint32_t i;

  if (argc != 3) {
    fprintf(stderr,"Uage: %s <store> <file>\n",argv[0]);
    return 1;
  }

  store_fd=open(argv[1],O_RDONLY);
  if (store_fd==-1) {
    perror(argv[1]);
    return 1;
  }

  store_size=size_in_blocks(store_fd);
  if (store_size) {
    switch (pread(store_fd,buf,512,0)) {
    case -1:
      perror(argv[1]);
      return 1;
    default:
      store_size=0;
    case 512:
      if (memcmp(buf,"TSTORE1",7)) store_size=0;
    }
  }
  if (!store_size) {
    fprintf(stderr,"%s: Not a valid store\n",argv[1]);
    return 1;
  }

  for (idx=store_size-1; idx; idx=nidx) {
    switch(pread(store_fd,buf,512,idx*512)) {
    case -1:
      perror(argv[1]);
      return 1;
    default:
      fprintf(stderr,"%s: short pread\n",argv[1]);
      return 1;
    case 512:
      break;
    }
    if (!strcmp(argv[2],buf+16)) {
      uint64_t size;
      uint32_t directory;
      memcpy(&size,buf+8,8);
      memcpy(&directory,buf+4,4);

      for (i=0;i<(size+511)/512;++i) {
	int j;
	int siz;
	uint32_t block=directory;
	for (j=3;j>=0;--j) {
	  get_block(block,(void*)pointers[j]);
	  block=pointers[j][(i>>(j*7))&127];
	}
	get_block(block,buf);
	siz=512;
	if (i==size/512) siz=size%512;
	assert(write(1,buf,siz)==siz);
      }
      return 0;
    }
    memcpy(&nidx,buf,4);
    if (nidx>=idx) {
      fprintf(stderr,"Invalid link %d=>%d\n",idx,nidx);
      return 1;
    }
  }
  fprintf(stderr,"%s: not found\n",argv[2]);
  return 1;
}
